@@ -30,7 +30,7 @@
 #include <windows.h>
 #define PATH_MAX 260
 #define PACKAGE "cava"
-#define VERSION "0.10.6"
+#define VERSION "0.10.7"
 #define _CRT_SECURE_NO_WARNINGS 1
 #endif // _WIN32
 
@@ -44,7 +44,6 @@
 
 #include "config.h"
 
-#include "debug.h"
 #include "util.h"
 
 #ifdef SDL
@@ -118,16 +117,31 @@ static int getopt(int argc, char *const argv[], const char *optstring) {
 // needs to know output mode in order to clean up terminal
 int output_mode;
 // whether we should reload the config or not
-int should_reload = 0;
+volatile sig_atomic_t should_reload = 0;
 // whether we should only reload colors or not
-int reload_colors = 0;
+volatile sig_atomic_t reload_colors = 0;
 // whether we should quit
-int should_quit = 0;
-int signal_received = 0;
+volatile sig_atomic_t should_quit = 0;
+volatile sig_atomic_t signal_received = 0;
+
+static bool get_file_state(const char *path, time_t *mtime, long long *size) {
+#ifdef _WIN32
+    struct _stat st;
+    if (_stat(path, &st) != 0)
+        return false;
+#else
+    struct stat st;
+    if (stat(path, &st) != 0)
+        return false;
+#endif
+    *mtime = st.st_mtime;
+    *size = (long long)st.st_size;
+    return true;
+}
 
 // these variables are used only in main, but making them global
 // will allow us to not free them on exit without ASan complaining
-struct config_params p;
+struct config_params p = {0};
 
 // general: cleanup
 void cleanup(void) {
@@ -315,20 +329,25 @@ Keys:\n\
             abort();
         }
     }
-    int reload = 0;
 
     // general: main loop
     while (1) {
 
-        debug("loading config\n");
         // config: load
         struct error_s error;
         error.length = 0;
-        if (!load_config(configPath, &p, 0, &error, reload)) {
+        if (!load_config(configPath, &p, &error)) {
             fprintf(stderr, "Error loading config. %s", error.message);
             exit(EXIT_FAILURE);
         }
 
+        time_t config_mtime = 0;
+        long long config_size = 0;
+        bool has_config_state = false;
+        if (p.live_config) {
+            has_config_state = get_file_state(configPath, &config_mtime, &config_size);
+        }
+
         int inAtty = 0;
         int inAterminal = 0;
 
@@ -348,7 +367,7 @@ Keys:\n\
                 strncmp(ttyname(0), "/dev/ttyUSB", 11) == 0)
                 inAterminal = 1;
 
-            // in macos vitual terminals are called ttys(xyz) and there are no ttys
+            // in macos virtual terminals are called ttys(xyz) and there are no ttys
             if (strncmp(ttyname(0), "/dev/ttys", 9) == 0)
                 inAtty = 0;
 
@@ -434,13 +453,11 @@ Keys:\n\
                                         // adjusted later if sample rate is unusual
 
         audio.cava_in = (double *)malloc(audio.cava_buffer_size * sizeof(double));
-        memset(audio.cava_in, 0, sizeof(int) * audio.cava_buffer_size);
+        memset(audio.cava_in, 0, sizeof(double) * audio.cava_buffer_size);
 
         audio.threadparams = 0; // most input threads don't adjust the parameters
         audio.terminate = 0;
 
-        debug("starting audio thread\n");
-
         pthread_t p_thread;
         int timeout_counter = 0;
         int total_bar_height = 0;
@@ -535,7 +552,7 @@ Keys:\n\
             audio.channels = p.channels;
             audio.active = p.active;
             audio.remix = p.remix;
-            audio.virtual = p.virtual;
+            audio.virtual_node = p.virtual_node;
             thr_id = pthread_create(&p_thread, NULL, input_pipewire, (void *)&audio);
             break;
 #endif
@@ -570,7 +587,6 @@ Keys:\n\
             }
         }
         pthread_mutex_unlock(&audio.lock);
-        debug("got format: %d and rate %d\n", audio.format, audio.rate);
 
         int audio_channels = audio.channels;
 
@@ -594,13 +610,21 @@ Keys:\n\
         int *bars;
         int *previous_frame;
 
+        int *right_bars = NULL;
+        int *right_previous_frame = NULL;
+
         float *bars_left, *bars_right;
         double *cava_out;
         float *bars_raw;
         float *previous_bars_raw;
 
-        int height, lines, width, remainder, fp;
+        int height, lines, width, remainder;
         int *dimension_bar, *dimension_value;
+#ifndef _WIN32
+        int fp = -1;
+        int keepalive_fd = -1;
+        bool close_output_fd = false;
+#endif
 #ifdef _WIN32
         HANDLE hFile = NULL;
 #endif
@@ -677,13 +701,21 @@ Keys:\n\
             case OUTPUT_NORITAKE:
                 if (strcmp(p.raw_target, "/dev/stdout") != 0) {
 #ifndef _WIN32
-                    int fptest;
+                    if (close_output_fd && fp != -1) {
+                        close(fp);
+                        fp = -1;
+                        close_output_fd = false;
+                    }
+                    if (keepalive_fd != -1) {
+                        close(keepalive_fd);
+                        keepalive_fd = -1;
+                    }
                     // checking if file exists
                     if (access(p.raw_target, F_OK) != -1) {
                         // file exists, testopening in case it's a fifo
-                        fptest = open(p.raw_target, O_RDONLY | O_NONBLOCK, 0644);
+                        keepalive_fd = open(p.raw_target, O_RDONLY | O_NONBLOCK, 0644);
 
-                        if (fptest == -1) {
+                        if (keepalive_fd == -1) {
                             fprintf(stderr, "could not open file %s for writing\n", p.raw_target);
                             exit(1);
                         }
@@ -694,9 +726,14 @@ Keys:\n\
                             exit(1);
                         }
                         // fifo needs to be open for reading in order to write to it
-                        fptest = open(p.raw_target, O_RDONLY | O_NONBLOCK, 0644);
+                        keepalive_fd = open(p.raw_target, O_RDONLY | O_NONBLOCK, 0644);
+                        if (keepalive_fd == -1) {
+                            fprintf(stderr, "could not open file %s for writing\n", p.raw_target);
+                            exit(1);
+                        }
                     }
                     fp = open(p.raw_target, O_WRONLY | O_NONBLOCK | O_CREAT, 0644);
+                    close_output_fd = true;
 #else
                     int pipeLength = strlen("\\\\.\\pipe\\") + strlen(p.raw_target) + 1;
                     char *pipePath = malloc(pipeLength);
@@ -727,10 +764,6 @@ Keys:\n\
                     exit(1);
                 }
 
-#ifndef NDEBUG
-                debug("open file %s for writing raw output\n", p.raw_target);
-#endif
-
                 // width must be hardcoded for raw output. only used to calculate the number of
                 // bars in auto mode
                 width = 512 * output_channels;
@@ -761,9 +794,10 @@ Keys:\n\
                     audio.terminate = 1;
                     exit(EXIT_FAILURE);
                 }
-                if (number_of_bars <= 1) {
+                if (number_of_bars < output_channels) {
                     cleanup();
-                    fprintf(stderr, "fixed number of bars must be at least 1\n");
+                    fprintf(stderr, "fixed number of bars must be at least 1 with mono output and "
+                                    "2 with stereo output\n");
                     exit(EXIT_FAILURE);
                 }
                 if (output_channels == 2 && number_of_bars % 2 != 0) {
@@ -776,11 +810,11 @@ Keys:\n\
 
                 if (output_mode == OUTPUT_SDL_GLSL) {
                     if (number_of_bars > 512)
-                        number_of_bars = 512; // cant have more than 512 bars in glsl due to shader
+                        number_of_bars = 512; // can't have more than 512 bars in glsl due to shader
                                               // program implementation limitations
                 } else {
                     if (number_of_bars / output_channels > 512)
-                        number_of_bars = 512 * output_channels; // cant have more than 512 bars on
+                        number_of_bars = 512 * output_channels; // can't have more than 512 bars on
                                                                 // 44100 rate per channel
                 }
 
@@ -794,12 +828,7 @@ Keys:\n\
                     number_of_bars--; // stereo must have even numbers of bars
             }
 
-            int raw_number_of_bars = (number_of_bars / output_channels) * audio_channels;
-            if (p.waveform) {
-                raw_number_of_bars = number_of_bars;
-            }
-
-            // checks if there is stil extra room, will use this to center
+            // checks if there is still extra room, will use this to center
             if (p.center_align) {
                 remainder = (*dimension_bar - number_of_bars * p.bar_width -
                              number_of_bars * p.bar_spacing + p.bar_spacing) /
@@ -817,14 +846,14 @@ Keys:\n\
                                         number_of_bars, width, lines, p.bar_width, p.orientation,
                                         p.blendDirection);
             }
+            if (p.horizontal_stereo) {
+                number_of_bars *= 2;
+            }
 
-#ifndef NDEBUG
-            debug("height: %d width: %d dimension_bar: %d dimension_value: %d bars:%d bar width: "
-                  "%d remainder: %d\n",
-                  height, width, *dimension_bar, *dimension_value, number_of_bars, p.bar_width,
-                  remainder);
-#endif
-
+            int raw_number_of_bars = (number_of_bars / output_channels) * audio_channels;
+            if (p.waveform) {
+                raw_number_of_bars = number_of_bars;
+            }
             double userEQ_keys_to_bars_ratio;
 
             if (p.userEQ_enabled && (number_of_bars / output_channels > 0)) {
@@ -871,6 +900,13 @@ Keys:\n\
             memset(previous_frame, 0, sizeof(int) * number_of_bars);
             memset(cava_out, 0, sizeof(double) * number_of_bars * audio.channels / output_channels);
 
+            if (p.horizontal_stereo) {
+                right_bars = (int *)malloc(number_of_bars * sizeof(int));
+                right_previous_frame = (int *)malloc(number_of_bars * sizeof(int));
+                memset(right_bars, 0, sizeof(int) * number_of_bars);
+                memset(right_previous_frame, 0, sizeof(int) * number_of_bars);
+            }
+
             int x_axis_info = 0;
             // process: calculate x axis values
             if (p.xaxis != NONE) {
@@ -923,14 +959,20 @@ Keys:\n\
 
             bool resizeTerminal = false;
 
-            int frame_time_msec = (1 / (float)p.framerate) * 1000;
-            struct timespec framerate_timer = {.tv_sec = 0, .tv_nsec = 0};
-            if (p.framerate <= 1) {
-                framerate_timer.tv_sec = frame_time_msec / 1000;
-            } else {
-                framerate_timer.tv_sec = 0;
-                framerate_timer.tv_nsec = frame_time_msec * 1e6;
-            }
+            int effective_framerate = p.framerate;
+            if (effective_framerate <= 0)
+                effective_framerate = 60;
+
+            long long frame_time_ns = (long long)(1000000000.0 / (double)effective_framerate);
+            if (frame_time_ns < 1)
+                frame_time_ns = 1;
+
+            int frame_time_msec = (int)(frame_time_ns / 1000000LL);
+            if (frame_time_msec < 1)
+                frame_time_msec = 1;
+
+            struct timespec framerate_timer = {.tv_sec = frame_time_ns / 1000000000LL,
+                                               .tv_nsec = frame_time_ns % 1000000000LL};
 #ifdef _WIN32
             LARGE_INTEGER frequency; // ticks per second
             LARGE_INTEGER t1, t2;    // ticks
@@ -942,11 +984,6 @@ Keys:\n\
             bool silence = false;
             char ch = '\0';
 
-#ifndef NDEBUG
-            int maxvalue = 0;
-            int minvalue = 0;
-#endif
-
             struct timespec sleep_mode_timer = {.tv_sec = 1, .tv_nsec = 0};
 
             int total_frames = 0;
@@ -986,14 +1023,14 @@ Keys:\n\
                 case 'c': // reload colors
                     reload_colors = 1;
                     break;
-                case 'f': // change forground color
+                case 'f': // change foreground color
                     if (p.col < 7)
                         p.col++;
                     else
                         p.col = 0;
                     resizeTerminal = true;
                     break;
-                case 'b': // change backround color
+                case 'b': // change background color
                     if (p.bgcol < 7)
                         p.bgcol++;
                     else
@@ -1035,7 +1072,6 @@ Keys:\n\
                 ch = 0;
 
                 if (should_reload) {
-
                     reloadConf = true;
                     resizeTerminal = true;
                     should_reload = 0;
@@ -1043,22 +1079,33 @@ Keys:\n\
 
                 if (reload_colors) {
                     struct error_s error;
+                    char *themeFile;
                     error.length = 0;
-                    if (!load_config(configPath, (void *)&p, 1, &error, reload)) {
+                    bool result = get_themeFile(configPath, &p, NULL, &error, &themeFile);
+                    if (!result) {
                         cleanup();
+                        exit(EXIT_FAILURE);
+                    }
+                    if (!load_colors(themeFile, (void *)&p, &error)) {
+                        cleanup();
+                        free(themeFile);
                         fprintf(stderr, "Error loading config. %s", error.message);
                         exit(EXIT_FAILURE);
                     }
                     resizeTerminal = true;
                     reload_colors = 0;
+                    free(themeFile);
+                    break;
                 }
 
-#ifndef NDEBUG
-                // clear();
-#ifndef _WIN32
-                refresh();
-#endif
-#endif
+                // checking if audio thread has exited unexpectedly
+                pthread_mutex_lock(&audio.lock);
+                if (audio.terminate == 1) {
+                    cleanup();
+                    fprintf(stderr, "Audio thread exited unexpectedly. %s\n", audio.error_message);
+                    exit(EXIT_FAILURE);
+                }
+                pthread_mutex_unlock(&audio.lock);
 
                 // process: check if input is present
                 silence = true;
@@ -1073,16 +1120,29 @@ Keys:\n\
 
                 if (output_mode != OUTPUT_SDL) {
                     if (p.sleep_timer) {
-                        if (silence && sleep_counter <= p.framerate * p.sleep_timer)
+                        int effective_framerate = p.framerate;
+                        if (effective_framerate <= 0)
+                            effective_framerate = 60;
+                        int sleep_limit = effective_framerate * p.sleep_timer;
+                        if (silence && sleep_counter <= sleep_limit)
                             sleep_counter++;
                         else if (!silence)
                             sleep_counter = 0;
 
-                        if (sleep_counter > p.framerate * p.sleep_timer) {
-#ifndef NDEBUG
-                            printw("no sound detected for 30 sec, going to sleep mode\n");
-#endif
+                        if (sleep_counter > sleep_limit) {
                             nanosleep(&sleep_mode_timer, NULL);
+                            if (p.live_config) {
+                                time_t new_mtime = 0;
+                                long long new_size = 0;
+                                if (get_file_state(configPath, &new_mtime, &new_size) &&
+                                    (!has_config_state || new_mtime != config_mtime ||
+                                     new_size != config_size)) {
+                                    should_reload = 1;
+                                    config_mtime = new_mtime;
+                                    config_size = new_size;
+                                    has_config_state = true;
+                                }
+                            }
                             continue;
                         }
                     }
@@ -1183,22 +1243,42 @@ Keys:\n\
                     if (audio_channels == 2) {
                         if (p.stereo) {
                             // mirroring stereo channels
-                            for (int n = 0; n < number_of_bars; n++) {
-                                if (n < number_of_bars / 2) {
-                                    if (p.reverse) {
-                                        bars_raw[n] = bars_left[n];
+                            if (p.horizontal_stereo) {
+                                for (int n = 0; n < number_of_bars; n++) {
+                                    if (n < number_of_bars / 2) {
+                                        if (p.reverse) {
+                                            bars_raw[n] = bars_left[number_of_bars / 2 - n - 1];
+                                        } else {
+                                            bars_raw[n] = bars_left[n];
+                                        }
                                     } else {
-                                        bars_raw[n] = bars_left[number_of_bars / 2 - n - 1];
+                                        if (p.reverse) {
+                                            bars_raw[n] = bars_right[number_of_bars - n - 1];
+                                        } else {
+                                            bars_raw[n] = bars_right[n - number_of_bars / 2];
+                                        }
                                     }
-                                } else {
-                                    if (p.reverse) {
-                                        bars_raw[n] = bars_right[number_of_bars - n - 1];
+                                }
+
+                            } else {
+                                for (int n = 0; n < number_of_bars; n++) {
+                                    if (n < number_of_bars / 2) {
+                                        if (p.reverse) {
+                                            bars_raw[n] = bars_left[n];
+                                        } else {
+                                            bars_raw[n] = bars_left[number_of_bars / 2 - n - 1];
+                                        }
                                     } else {
-                                        bars_raw[n] = bars_right[n - number_of_bars / 2];
+                                        if (p.reverse) {
+                                            bars_raw[n] = bars_right[number_of_bars - n - 1];
+                                        } else {
+                                            bars_raw[n] = bars_right[n - number_of_bars / 2];
+                                        }
                                     }
                                 }
                             }
                         } else {
+
                             // stereo mono output
                             for (int n = 0; n < number_of_bars; n++) {
                                 if (p.reverse) {
@@ -1240,43 +1320,10 @@ Keys:\n\
 
                     if (bars[n] != previous_frame[n])
                         re_paint = 1;
-#endif
-
-#ifndef NDEBUG
-                    mvprintw(n, 0, "%d: f:%f->%f (%d->%d), eq:\
-						%15e, peak:%d \n",
-                             n, plan->cut_off_frequency[n], plan->cut_off_frequency[n + 1],
-                             plan->FFTbuffer_lower_cut_off[n], plan->FFTbuffer_upper_cut_off[n],
-                             plan->eq[n], bars[n]);
-
-                    if (bars[n] < minvalue) {
-                        minvalue = bars[n];
-                        debug("min value: %d\n", minvalue); // checking maxvalue 10000
-                    }
-                    if (bars[n] > maxvalue) {
-                        maxvalue = bars[n];
-                    }
-                    if (bars[n] < 0) {
-                        debug("negative bar value!! %d\n", bars[n]);
-                        //    exit(EXIT_FAILURE); // Can't happen.
-                    }
-
 #endif
                 }
 
-#ifndef NDEBUG
-                mvprintw(number_of_bars + 1, 0, "sensitivity %.10e", p.sens);
-                mvprintw(number_of_bars + 2, 0, "min value: %d\n",
-                         minvalue); // checking maxvalue 10000
-                mvprintw(number_of_bars + 3, 0, "max value: %d\n",
-                         maxvalue); // checking maxvalue 10000
-#ifndef _WIN32
-                (void)x_axis_info;
-#endif // !_WIN32
-#endif
-
-// output: draw processed input
-#ifdef NDEBUG
+                // output: draw processed input
                 if (p.sync_updates) {
                     printf("\033[2026h\033\\");
                     fflush(stdout);
@@ -1303,14 +1350,39 @@ Keys:\n\
 #endif
                 case OUTPUT_NONCURSES:
                     if (p.orientation == ORIENT_SPLIT_H) {
-                        rc = draw_terminal_noncurses(
-                            inAtty, lines, width, number_of_bars, p.bar_width, p.bar_spacing,
-                            remainder, bars, previous_frame, p.gradient, p.horizontal_gradient,
-                            x_axis_info, ORIENT_BOTTOM, 1);
-                        rc = draw_terminal_noncurses(
-                            inAtty, lines, width, number_of_bars, p.bar_width, p.bar_spacing,
-                            remainder, bars, previous_frame, p.gradient, p.horizontal_gradient,
-                            x_axis_info, ORIENT_TOP, 1);
+                        // in split horizontal mode we need to draw two times
+                        // once for bottom and once for top
+                        // ORIENT_BOTTOM will be the top bars and ORIENT_TOP the bottom bars
+                        // since bottom hear means from mid upwards and top is from mid downwards
+                        if (p.horizontal_stereo) {
+                            // in horizontal stereo mode we need to split the bars array in half
+                            // first half is right channel, second half is left channel
+                            for (int i = 0; i < number_of_bars / 2; i++) {
+                                right_bars[i] = bars[i + number_of_bars / 2];
+                                right_previous_frame[i] = previous_frame[i + number_of_bars / 2];
+                            }
+
+                            rc = draw_terminal_noncurses(inAtty, lines, width, number_of_bars / 2,
+                                                         p.bar_width, p.bar_spacing, remainder,
+                                                         right_bars, right_previous_frame,
+                                                         p.gradient, p.horizontal_gradient,
+                                                         x_axis_info, 1 - p.left_bottom, 1);
+
+                            rc = draw_terminal_noncurses(
+                                inAtty, lines, width, number_of_bars / 2, p.bar_width,
+                                p.bar_spacing, remainder, bars, previous_frame, p.gradient,
+                                p.horizontal_gradient, x_axis_info, p.left_bottom, 1);
+
+                        } else {
+                            rc = draw_terminal_noncurses(
+                                inAtty, lines, width, number_of_bars, p.bar_width, p.bar_spacing,
+                                remainder, bars, previous_frame, p.gradient, p.horizontal_gradient,
+                                x_axis_info, ORIENT_BOTTOM, 1);
+                            rc = draw_terminal_noncurses(
+                                inAtty, lines, width, number_of_bars, p.bar_width, p.bar_spacing,
+                                remainder, bars, previous_frame, p.gradient, p.horizontal_gradient,
+                                x_axis_info, ORIENT_TOP, 1);
+                        }
                     } else {
                         rc = draw_terminal_noncurses(
                             inAtty, lines, width, number_of_bars, p.bar_width, p.bar_spacing,
@@ -1363,21 +1435,11 @@ Keys:\n\
                     should_quit = true;
                 }
 
-#endif
-
                 memcpy(previous_frame, bars, number_of_bars * sizeof(int));
                 if (p.output == OUTPUT_SDL_GLSL) {
                     memcpy(previous_bars_raw, bars_raw, number_of_bars * sizeof(float));
                 }
 
-                // checking if audio thread has exited unexpectedly
-                pthread_mutex_lock(&audio.lock);
-                if (audio.terminate == 1) {
-                    cleanup();
-                    fprintf(stderr, "Audio thread exited unexpectedly. %s\n", audio.error_message);
-                    exit(EXIT_FAILURE);
-                }
-                pthread_mutex_unlock(&audio.lock);
 #ifdef _WIN32
                 QueryPerformanceCounter(&t2);
                 elapsedTime = (t2.QuadPart - t1.QuadPart) * 1000.0 / frequency.QuadPart;
@@ -1397,6 +1459,19 @@ Keys:\n\
 #endif
                 }
 
+                if (p.live_config) {
+                    time_t new_mtime = 0;
+                    long long new_size = 0;
+                    if (get_file_state(configPath, &new_mtime, &new_size) &&
+                        (!has_config_state || new_mtime != config_mtime ||
+                         new_size != config_size)) {
+                        should_reload = 1;
+                        config_mtime = new_mtime;
+                        config_size = new_size;
+                        has_config_state = true;
+                    }
+                }
+
                 if (p.draw_and_quit > 0) {
                     total_frames++;
                     if (total_frames >= p.draw_and_quit) {
@@ -1425,6 +1500,25 @@ Keys:\n\
             free(bars_raw);
             free(previous_bars_raw);
             free(previous_frame);
+            if (p.horizontal_stereo) {
+                free(right_bars);
+                free(right_previous_frame);
+            }
+
+#ifndef _WIN32
+            if ((output_mode == OUTPUT_RAW || output_mode == OUTPUT_NORITAKE) &&
+                strcmp(p.raw_target, "/dev/stdout") != 0) {
+                if (close_output_fd && fp != -1) {
+                    close(fp);
+                    fp = -1;
+                    close_output_fd = false;
+                }
+                if (keepalive_fd != -1) {
+                    close(keepalive_fd);
+                    keepalive_fd = -1;
+                }
+            }
+#endif
         } // reloading config
 
         //**telling audio thread to terminate**//
@@ -1433,13 +1527,11 @@ Keys:\n\
         pthread_mutex_unlock(&audio.lock);
         pthread_join(p_thread, NULL);
 
-        if (p.userEQ_enabled)
-            free(p.userEQ);
+        pthread_mutex_destroy(&audio.lock);
 
         free(audio.source);
         free(audio.cava_in);
         cleanup();
-        reload = 1;
 
         if (should_quit && signal_received == 0) {
             if (p.zero_test && total_bar_height > 0) {
